"""Foot warmer representation for SleepIQ API."""
from __future__ import annotations

from .api import SleepIQAPI
from .consts import (Side, FootWarmingTemps, SIDES_FULL)

class SleepIQFootWarmer:
    """Foot warmer representation for SleepIQ API."""

    def __init__(self, api: SleepIQAPI, bed_id: str, side: Side, timer: int, temperature: int) -> None:
        """Initialize foot warmer object."""
        self._api = api
        self.bed_id = bed_id
        self.side = side
        self.is_on = (temperature > 0)
        self.timer = timer
        self.temperature = temperature

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQFootWarmer[{self.side}]: {'On' if self.is_on else 'Off'}, {self.timer}, {FootWarmingTemps(self.temperature).name}"
    __repr__ = __str__

    async def turn_on(self, temperature: FootWarmingTemps, time: int)  -> None:
       """Turn on foot warmer through API."""
       await self.set_foot_warming(temperature, time)

    async def turn_off(self) -> None:
        """Turn off foot warmer through API."""
        # The API requires a valid time value even if we're turning the warmer off
        await self.set_foot_warming(FootWarmingTemps.OFF, 1)

    async def set_foot_warming(self, temperature: FootWarmingTemps, time: int) -> None:
        """Set foot warmer state through API."""
        if time <= 0 or time > 360:
            raise ValueError("Invalid Time, must be between 0 and 360")

        data = { self.time_key(): time, self.temp_key(): temperature.value}
        await self._api.put(f"bed/{self.bed_id}/foundation/footwarming", data)
        await self.update(data)

    async def update(self, data) -> None:
        # when reading the values the key is footWarmingStatus and when writing the values the key is footWarmingTemp so lookup both
        self.temperature = data.get(self.temp_key(True), data.get(self.temp_key()))
        self.is_on = self.temperature > 0
        self.timer = data.get(self.time_key(), self.timer) if self.is_on else 0

    def time_key(self):
        return 'footWarmingTimer' + SIDES_FULL[self.side]

    def temp_key(self, read = False):
        key = 'Status' if read else 'Temp'
        return f'footWarming{key}' + SIDES_FULL[self.side]



"""Sleeper representation for SleepIQ API."""
from __future__ import annotations

from .api import SleepIQAPI
from .consts import SIDES_FULL, SIDES_SHORT, Side


class SleepIQSleeper:
    """Sleeper representation for SleepIQ API."""

    def __init__(
        self, api: SleepIQAPI, bed_id: str, sleeper_id: str, side: Side
    ) -> None:
        """Initialize sleeper object."""
        self.api = api
        self.bed_id = bed_id
        self.sleeper_id = sleeper_id
        self.side = side
        self.side_full = SIDES_FULL[side]
        self.active = False
        self.name = ""

        self.in_bed = False
        self.pressure = 0
        self.sleep_number = 0
        self.fav_sleep_number = 0

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQSleeper[{self.side}]({self.name}, in_bed={self.in_bed}, sn={self.sleep_number})"

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SleepIQSleeper[{self.side}]({self.name}, in_bed={self.in_bed}, sn={self.sleep_number})"
    
    async def update(self) -> None:
        """Updates sleeper with latest data."""
        pass

    async def calibrate(self) -> None:
        """Calibrate or "baseline" bed."""
        await self.api.put("sleeper/" + self.sleeper_id + "/calibrate")

    async def set_sleepnumber(self, setting: int) -> None:
        """Set sleep number 5-100 (multiple of 5)."""
        if 0 > setting or setting > 100:
            raise ValueError("Invalid SleepNumber, must be between 0 and 100")
        setting = int(round(setting / 5)) * 5
        data = {
            "sleepNumber": setting, 
            "side": SIDES_SHORT[self.side],
        }
        await self.api.put("bed/" + self.bed_id + "/sleepNumber", data)

    async def set_favsleepnumber(self, setting: int) -> None:
        """Set favorite sleep number 5-100 (multiple of 5)."""
        if 0 > setting or setting > 100:
            raise ValueError("Invalid SleepNumber, must be between 0 and 100")
        setting = int(round(setting / 5)) * 5
        data = {
            "side": SIDES_SHORT[self.side],
            "sleepNumberFavorite": setting,
        }
        await self.api.put("bed/" + self.bed_id + "/sleepNumberFavorite", data)
        await self.fetch_favsleepnumber()

    async def fetch_favsleepnumber(self) -> None:
        """Update fav_sleep_number from API."""
        json = await self.api.get("bed/" + self.bed_id + "/sleepNumberFavorite")
        self.fav_sleep_number = json["sleepNumberFavorite" + self.side_full]



"""Foundation preset setting from SleepIQ API."""
from __future__ import annotations

from typing import Any

from ..preset import SleepIQPreset
from ..api import SleepIQAPI
from ..consts import (
    NO_PRESET,
    PRESET_FAV,
    PRESET_READ,
    PRESET_TV,
    PRESET_FLAT,
    PRESET_ZERO_G,
    PRESET_SNORE,
    Side,
)

PRESET_VALS = {
    PRESET_FAV: "favorite",
    PRESET_READ: "read",
    PRESET_TV: "watch_tv",
    PRESET_FLAT: "flat",
    PRESET_ZERO_G: "zero_g",
    PRESET_SNORE: "snore",
}


class SleepIQFuzionPreset(SleepIQPreset):
    """Foundation preset setting from SleepIQ API."""

    def __init__(self, api: SleepIQAPI, bed_id: str, side: Side, options: list[str]) -> None:
        """Initialize preset setting."""
        super().__init__(api, bed_id, side)
        self.options = options

    async def set_preset(self, preset: str, slow_speed: bool = False) -> None:
        """Set foundation preset."""
        if preset == NO_PRESET:
            return
        if preset not in self.options:
            raise ValueError("Invalid preset")
        args = [self.side_full.lower(), PRESET_VALS[preset]]
        await self._api.bamkey(self.bed_id, "SetTargetPresetWithoutTimer", args)

    async def update(self, data: dict[str, Any]) -> None:
        """Update the position of an actuator from the API."""
        args = [self.side_full.lower()]
        self.preset = await self._api.bamkey(self.bed_id, "GetCurrentPreset", args)



"""SleepIQ Exceptions."""


class SleepIQLoginException(Exception):
    """Exception in Login process."""


class SleepIQTimeoutException(Exception):
    """Timeout in Login process."""


class SleepIQAPIException(Exception):
    """Exception in API call."""

    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__(message)



"""Actuator representation for SleepIQ API."""
from __future__ import annotations

from typing import Any

from .api import SleepIQAPI
from .consts import ACTUATORS_FULL, SIDES_FULL, SIDES_SHORT, End, Side


class SleepIQActuator:
    """Actuator representation for SleepIQ API."""

    def __init__(
        self, api: SleepIQAPI, bed_id: str, side: Side, actuator: End
    ) -> None:
        """Initialize actuator object."""
        self._api = api
        self.bed_id = bed_id
        self.side = side
        self.side_full = SIDES_FULL[side]
        self.actuator = actuator
        self.actuator_full = ACTUATORS_FULL[actuator]
        self.position = 0

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQActuator[{self.actuator_full} {self.side}], position={self.position}"

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SleepIQActuator[{self.actuator_full} {self.side}], position={self.position}"

    async def set_position(self, position: int, slow_speed: bool = False) -> None:
        """Set the position of an actuator through the API."""
        if position < 0 or position > 100:
            raise ValueError("Invalid position, must be between 0 and 100")
        if position == self.position:
            return
        data = {
            "position": position,
            "side": SIDES_SHORT[self.side],
            "actuator": self.actuator,
            "speed": 1 if slow_speed else 0,
        }
        await self._api.put(f"bed/{self.bed_id}/foundation/adjustment/micro", data)

    async def update(self, data: dict[str, Any]) -> None:
        """Update the position of an actuator from the API."""
        # The API reports position in hex, but is set with an integer.
        # We'll always show position with an integer value.
        self.position = int(data[f"fs{self.side_full}{self.actuator_full}Position"], 16)



"""Bed object from SleepIQ API."""
from __future__ import annotations

from typing import Any

from .api import SleepIQAPI
from .consts import SIDES_FULL, Side
from .foundation import SleepIQFoundation
from .sleeper import SleepIQSleeper


class SleepIQBed:
    """Bed object from SleepIQ API."""

    def __init__(self, api: SleepIQAPI, data: dict[str, Any]) -> None:
        """Initialize bed object."""
        self._api = api

        self.name = data["name"]
        self.id = data["bedId"]
        self.mac_addr = data["macAddress"]
        self.paused = False
        self.sleepers = [
            SleepIQSleeper(api, self.id, data[f"sleeper{SIDES_FULL[side]}Id"], side)
            for side in [Side.LEFT, Side.RIGHT]
            if data.get(f"sleeper{SIDES_FULL[side]}Id")
        ]
        self.foundation = SleepIQFoundation(api, self.id)

        self.model = "Unknown"
        if "model" in data:
            self.model = data["model"]
        elif "components" in data:
            for comp in data["components"]:
                if comp["base"] == "BASE":
                    self.model = comp["model"]
                    break

    def __str__(self) -> str:
        """Return string representation."""
        return (
            f"SleepIQBed({self.name}, model={self.model}, id={self.id}) "
            + str(self.sleepers)
            + " "
            + str(self.foundation)
        )

    def __repr__(self) -> str:
        """Return string representation."""
        return (
            f"SleepIQBed({self.name}, model={self.model}, id={self.id}) "
            + str(self.sleepers)
            + " "
            + str(self.foundation)
        )

    async def valid(self) -> bool:
        return await self._api.check("bed/" + self.id + "/pauseMode")

    async def calibrate(self) -> None:
        """Calibrate or "baseline" bed."""
        for sleeper in self.sleepers:
            if sleeper.sleeper_id:
                await sleeper.calibrate()
                break

    async def stop_pump(self) -> None:
        """Stop pump."""
        await self._api.put("bed/" + self.id + "/pump/forceIdle")

    async def fetch_pause_mode(self) -> None:
        """Update paused attribute with data from API."""
        json = await self._api.get("bed/" + self.id + "/pauseMode")
        self.paused = json.get("pauseMode", "") == "on"

    async def set_pause_mode(self, mode: bool) -> None:
        """Set pause mode in API and locally."""
        params = {"mode": "on" if mode else "off"}
        await self._api.put("bed/" + self.id + "/pauseMode", params=params)
        self.paused = mode



"""Foundation preset setting from SleepIQ API."""
from __future__ import annotations

from typing import Any
from .api import SleepIQAPI
from .consts import BED_PRESETS, NO_PRESET, SIDES_FULL, SIDES_SHORT, Side


class SleepIQPreset:
    """Foundation preset setting from SleepIQ API."""

    def __init__(self, api: SleepIQAPI, bed_id: str, side: Side) -> None:
        """Initialize preset setting."""
        self._api = api
        self.bed_id = bed_id
        self.side = side
        self.side_full = SIDES_FULL[side]
        self.preset = ""
        self.options = list(BED_PRESETS)

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQPreset[{self.side}](preset={self.preset})"

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SleepIQPreset[{self.side}](preset={self.preset})"

    async def set_preset(self, preset: str, slow_speed: bool = False) -> None:
        """Set foundation preset."""
        #
        # preset 1-6
        # slowSpeed False=fast, True=slow
        #
        if preset == NO_PRESET:
            return
        if preset not in self.options:
            raise ValueError("Invalid preset")
        data = {"preset": BED_PRESETS[preset], "side": SIDES_SHORT[self.side], "speed": 1 if slow_speed else 0}
        await self._api.put("bed/" + self.bed_id + "/foundation/preset", data)

    async def update(self, data: dict[str, Any]) -> None:
        """Update the position of an actuator from the API."""
        self.preset = data[f"fsCurrentPositionPreset{self.side_full}"]



"""Foundation object from SleepIQ API."""
from __future__ import annotations

from typing import Any

from ..consts import (
    NO_PRESET,
    PRESET_FAV,
    PRESET_FLAT,
    PRESET_READ,
    PRESET_SNORE,
    PRESET_TV,
    PRESET_ZERO_G,
    End,
    Mode,
    Side,
    Speed,
)
from ..foundation import SleepIQFoundation
from .preset import SleepIQFuzionPreset
from .actuator import SleepIQFuzionActuator
from .light import SleepIQFuzionLight

FEATURE_NAMES = [
    "bedType", # Not sure what best to call this, but there's one flag at the start of the list that's (from testing) always "dual".
    "pressureControlEnabledFlag",
    "articulationEnableFlag",
    "underbedLightEnableFlag",
    "rapidSleepSettingEnableFlag",
    "thermalControlEnabledFlag",
    "rightHeadActuator",
    "rightFootActuator",
    "leftHeadActuator",
    "leftFootActuator",
    "flatPreset",
    "favoritePreset",
    "snorePreset",
    "zeroGravityPreset",
    "watchTvPreset",
    "readPreset",
]


class SleepIQFuzionFoundation(SleepIQFoundation):
    """Foundation object from SleepIQ API."""

    async def init_features(self) -> None:
        """Initialize all foundation features."""
        if self.features["underbedLightEnableFlag"]:
            await self.init_lights()
        if self.features["articulationEnableFlag"]:
            await self.init_actuators()
        await self.init_presets({})

    async def update_foundation_status(self) -> None:
        """Update all foundation data from API."""
        await self.update_lights()
        await self.update_actuators({})
        await self.update_presets({})

    async def init_lights(self) -> None:
        """Initialize list of lights available on foundation."""
        self.lights.append(SleepIQFuzionLight(self._api, self.bed_id, 1))

    async def init_actuators(self) -> None:
        """Initialize list of actuators available on foundation."""
        if self.features["rightHeadActuator"]:
            self.actuators.append(SleepIQFuzionActuator(self._api, self.bed_id, Side.RIGHT, End.HEAD))
        if self.features["rightFootActuator"]:
            self.actuators.append(SleepIQFuzionActuator(self._api, self.bed_id, Side.RIGHT, End.FOOT))
        if self.features["leftHeadActuator"]:
            self.actuators.append(SleepIQFuzionActuator(self._api, self.bed_id, Side.LEFT, End.HEAD))
        if self.features["leftFootActuator"]:
            self.actuators.append(SleepIQFuzionActuator(self._api, self.bed_id, Side.LEFT, End.FOOT))

        await self.update_actuators([])

    async def init_presets(self, data: dict[str, Any]) -> None:
        """Initialize list of presets available on foundation."""
        options = [NO_PRESET]
        if self.features["flatPreset"]:
            options.append(PRESET_FLAT)
        if self.features["favoritePreset"]:
            options.append(PRESET_FAV)
        if self.features["snorePreset"]:
            options.append(PRESET_SNORE)
        if self.features["zeroGravityPreset"]:
            options.append(PRESET_ZERO_G)
        if self.features["watchTvPreset"]:
            options.append(PRESET_TV)
        if self.features["readPreset"]:
            options.append(PRESET_READ)

        if self.type in ["single", "easternKing"]:
            self.presets = [SleepIQFuzionPreset(self._api, self.bed_id, Side.NONE, options)]
        else:
            self.presets = [
                SleepIQFuzionPreset(self._api, self.bed_id, Side.LEFT, options),
                SleepIQFuzionPreset(self._api, self.bed_id, Side.RIGHT, options),
            ]

        await self.update_presets({})

    async def fetch_features(self) -> None:
        """Update list of features available for foundation from API."""
        vals = await self._api.bamkey(self.bed_id, "GetSystemConfiguration")
        for k, v in zip(FEATURE_NAMES, vals.split()):
            if v == "no":
                v = False
            elif v == "yes":
                v = True
            self.features[k] = v

    async def stop_motion(self, side: str) -> None:
        """Stop motion on L or R side of bed."""
        await self._api.bamkey(self.bed_id, "HaltAllActuators")

    async def set_foundation_massage(
        self, side: str, foot_speed: Speed, head_speed: Speed, timer: int = 0, mode: Mode = Mode.OFF
    ) -> None:
        """Set massage mode."""
        pass



"""Foundation object from SleepIQ API."""
from __future__ import annotations

from typing import Any

from .actuator import SleepIQActuator
from .api import SleepIQAPI
from .consts import (
    BED_LIGHTS,
    FOUNDATION_TYPES,
    End,
    Mode,
    Side,
    Speed,
)
from .light import SleepIQLight
from .foot_warmer import SleepIQFootWarmer
from .preset import SleepIQPreset


class SleepIQFoundation:
    """Foundation object from SleepIQ API."""

    def __init__(self, api: SleepIQAPI, bed_id: str) -> None:
        """Initialize foundation object."""
        self._api = api
        self.bed_id = bed_id
        self.lights: list[SleepIQLight] = []
        self.foot_warmers: list[SleepIQFootWarmer] = []
        self.features: dict[str, bool] = {
            "boardIsASingle": False,
            "hasMassageAndLight": False,
            "hasFootControl": False,
            "hasFootWarming": False,
            "hasUnderbedLight": False,
            "leftUnderbedLightPMW": False,
            "rightUnderbedLightPMW": False,
        }
        self.type = ""
        self.actuators: list[SleepIQActuator] = []
        self.presets: list[SleepIQPreset] = []

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQFoundation[{self.type}](lights: {len(self.lights)}, features: {len(self.features)}, actuators: {len(self.actuators)}, presets: {len(self.presets)})"

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SleepIQFoundation[{self.type}](lights: {len(self.lights)}, features: {len(self.features)}, actuators: {len(self.actuators)}, presets: {len(self.presets)})"

    async def init_features(self) -> None:
        """Initialize all foundation features."""
        await self.init_lights()
        await self.init_foot_warmers()

        if not self.type:
            return

        data = await self._api.get(f"bed/{self.bed_id}/foundation/status")
        await self.init_actuators(data)
        await self.init_presets(data)

    async def update_foundation_status(self) -> None:
        """Update all foundation data from API."""
        await self.update_lights()
        await self.update_foot_warmers()

        if not self.type:
            return

        data = await self._api.get(f"bed/{self.bed_id}/foundation/status")

        await self.update_actuators(data)
        await self.update_presets(data)

    async def init_foot_warmers(self) -> None:
        if not self.features["hasFootWarming"]:
            return

        for side in [Side.LEFT, Side.RIGHT]:
            self.foot_warmers.append(SleepIQFootWarmer(self._api, self.bed_id, side, 0, 0))

    async def update_foot_warmers(self) -> None:
        if not self.features["hasFootWarming"]:
            return

        data = await self._api.get(f"bed/{self.bed_id}/foundation/footwarming")
        for foot_warmer in self.foot_warmers:
            await foot_warmer.update(data)

    async def init_lights(self) -> None:
        """Initialize list of lights available on foundation."""
        for light in BED_LIGHTS:
            exists = await self._api.check(f"bed/{self.bed_id}/foundation/outlet", params={"outletId": light})
            if exists:
                self.lights.append(SleepIQLight(self._api, self.bed_id, light))
        await self.update_lights()

    async def update_lights(self) -> None:
        """Update light states from API."""
        for light in self.lights:
            await light.update()

    async def init_actuators(self, data: dict[str, Any]) -> None:
        """Initialize list of actuators available on foundation."""

        if self.type in ["single", "easternKing"]:
            self.actuators = [
                SleepIQActuator(self._api, self.bed_id, Side.NONE, End.HEAD),
                SleepIQActuator(self._api, self.bed_id, Side.NONE, End.FOOT),
            ]
            self.presets = [SleepIQPreset(self._api, self.bed_id, Side.NONE)]
        elif self.type == "splitHead":
            self.actuators = [
                SleepIQActuator(self._api, self.bed_id, Side.LEFT, End.HEAD),
                SleepIQActuator(self._api, self.bed_id, Side.RIGHT, End.HEAD),
                SleepIQActuator(self._api, self.bed_id, Side.NONE, End.FOOT),
            ]
        else:
            self.actuators = [
                SleepIQActuator(self._api, self.bed_id, Side.LEFT, End.HEAD),
                SleepIQActuator(self._api, self.bed_id, Side.RIGHT, End.HEAD),
                SleepIQActuator(self._api, self.bed_id, Side.LEFT, End.FOOT),
                SleepIQActuator(self._api, self.bed_id, Side.RIGHT, End.FOOT),
            ]

        await self.update_actuators(data)

    async def init_presets(self, data: dict[str, Any]) -> None:
        """Initialize list of presets available on foundation."""
        if self.type in ["single", "easternKing"]:
            self.presets = [SleepIQPreset(self._api, self.bed_id, Side.NONE)]
        else:
            self.presets = [
                SleepIQPreset(self._api, self.bed_id, Side.LEFT),
                SleepIQPreset(self._api, self.bed_id, Side.RIGHT),
            ]

        await self.update_presets(data)

    async def update_actuators(self, data: dict[str, Any]) -> None:
        """Update actuator states from API."""
        for actuator in self.actuators:
            await actuator.update(data)

    async def update_presets(self, data: dict[str, Any]) -> None:
        """Update actuator states from API."""
        for preset in self.presets:
            await preset.update(data)

    async def fetch_features(self) -> None:
        """Update list of features available for foundation from API."""
        have_foundation = await self._api.check("bed/" + self.bed_id + "/foundation/system")
        if not have_foundation:
            self.type = ""
            return

        fs = await self._api.get("bed/" + self.bed_id + "/foundation/system")
        features_flags = fs.get("fsBoardFeatures", 0)
        self.features["boardIsASingle"] = bool(features_flags & (1 << 0))
        self.features["hasMassageAndLight"] = bool(features_flags & (1 << 1))
        self.features["hasFootControl"] = bool(features_flags & (1 << 2))
        self.features["hasFootWarming"] = bool(features_flags & (1 << 3))
        self.features["hasUnderbedLight"] = bool(features_flags & (1 << 4))
        type_num = int(fs.get("fsBedType", -1))
        if type_num < 0 or type_num > len(FOUNDATION_TYPES) - 1:
            self.type = ""
        else:
            self.type = FOUNDATION_TYPES[type_num]

        self.features["leftUnderbedLightPMW"] = bool(fs.get("fsLeftUnderbedLightPWM", False))
        self.features["rightUnderbedLightPMW"] = bool(fs.get("fsRightUnderbedLightPWM", False))

        if self.features["hasMassageAndLight"]:
            self.features["hasUnderbedLight"] = True
        if "split" in self.type:
            self.features["boardIsASingle"] = False

    async def stop_motion(self, side: str) -> None:
        """Stop motion on L or R side of bed."""
        data = {"footMotion": 1, "headMotion": 1, "massageMotion": 1, "side": side}
        await self._api.put("bed/" + self.bed_id + "/foundation/motion", data)

    async def set_foundation_massage(
        self, side: str, foot_speed: Speed, head_speed: Speed, timer: int = 0, mode: Mode = Mode.OFF
    ) -> None:
        """Set massage mode."""
        if mode != Mode.OFF:
            foot_speed = Speed.OFF
            head_speed = Speed.OFF

        data = {
            "footMassageMotor": foot_speed,
            "headMassageMotor": head_speed,
            "massageTimer": timer,
            "massageWaveMode": mode,
            "side": side,
        }
        await self._api.put("bed/" + self.bed_id + "/foundation/adjustment", data)



"""Sleeper representation for SleepIQ API."""
from __future__ import annotations

from ..consts import SIDES_FULL
from ..sleeper import SleepIQSleeper


class SleepIQFuzionSleeper(SleepIQSleeper):
    """Sleeper representation for SleepIQ API."""

    async def update(self) -> None:
        """Updates sleeper with latest data."""
        await self.fetch_sleepnumber()

    async def set_sleepnumber(self, setting: int) -> None:
        """Set sleep number 5-100 (multiple of 5)."""
        if 0 > setting or setting > 100:
            raise ValueError("Invalid SleepNumber, must be between 0 and 100")
        setting = int(round(setting / 5)) * 5
        args = [SIDES_FULL[self.side].lower(), str(setting)]
        await self.api.bamkey(self.bed_id, "StartSleepNumberAdjustment", args=args)

    async def fetch_sleepnumber(self) -> None:
        """Update fav_sleep_number from API."""
        args = [SIDES_FULL[self.side].lower()]
        result = await self.api.bamkey(self.bed_id, "GetSleepNumberControls", args=args)
        is_updating, ambient_number, user_number = result.split()
        self.sleep_number = int(user_number)

    async def set_favsleepnumber(self, setting: int) -> None:
        """Set favorite sleep number 5-100 (multiple of 5)."""
        if 0 > setting or setting > 100:
            raise ValueError("Invalid SleepNumber, must be between 0 and 100")
        setting = int(round(setting / 5)) * 5
        args = [SIDES_FULL[self.side].lower(), str(setting)]
        await self.api.bamkey(self.bed_id, "SetFavoriteSleepNumber", args=args)
        await self.fetch_favsleepnumber()

    async def fetch_favsleepnumber(self) -> None:
        """Update fav_sleep_number from API."""
        args = [SIDES_FULL[self.side].lower()]
        result = await self.api.bamkey(self.bed_id, "GetFavoriteSleepNumber", args=args)
        self.fav_sleep_number = int(result)



"""Light representation for SleepIQ API."""
from __future__ import annotations

from .api import SleepIQAPI


class SleepIQLight:
    """Light representation for SleepIQ API."""

    def __init__(self, api: SleepIQAPI, bed_id: str, outlet_id: int) -> None:
        """Initialize light object."""
        self._api = api
        self.bed_id = bed_id
        self.is_on = False
        self.outlet_id = outlet_id

    def __str__(self) -> str:
        """Return string representation."""
        return f"SleepIQLight[{self.outlet_id}]: {'On' if self.is_on else 'Off'}"

    def __repr__(self) -> str:
        """Return string representation."""
        return f"SleepIQLight[{self.outlet_id}]: {'On' if self.is_on else 'Off'}"

    async def turn_on(self) -> None:
        """Turn on light through API."""
        await self.set_light(True)
        self.is_on = True

    async def turn_off(self) -> None:
        """Turn off light through API."""
        await self.set_light(False)
        self.is_on = False

    async def set_light(self, setting: bool) -> None:
        """Set light state through API."""
        data = {"outletId": self.outlet_id, "setting": 1 if setting else 0}
        await self._api.put(f"bed/{self.bed_id}/foundation/outlet", data)

    async def update(self) -> None:
        """Update light state from API."""
        params = {"outletId": self.outlet_id}
        status = await self._api.get(
            f"bed/{self.bed_id}/foundation/outlet", params=params
        )
        self.is_on = status["setting"] == 1



"""AsyncSleepIQ class connects to the SleepIQ API and provides bed information."""
from __future__ import annotations

from aiohttp import ClientSession
import logging

from .api import SleepIQAPI
from .bed import SleepIQBed
from .consts import LOGIN_KEY
from .fuzion.bed import SleepIQFuzionBed
from .exceptions import SleepIQAPIException

_LOGGER = logging.getLogger("ASyncSleepIQ")


class AsyncSleepIQ(SleepIQAPI):
    """Representation of SleepIQ API object."""

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        login_method: int = LOGIN_KEY,
        client_session: ClientSession | None = None,
    ) -> None:
        """Initialize AsyncSleepIQ."""
        super().__init__(email, password, login_method, client_session)
        self.beds: dict[str, SleepIQBed] = {}

    # initialize beds and sleepers from API
    async def init_beds(self) -> None:
        """Initialize bed and sleeper objects from API data."""
        data = await self.get("bed")

        self._account_id = data["beds"][0].get("accountId", "")

        # get beds
        self.beds = {}
        for bed_data in data["beds"]:
            try:
                if bed_data.get("generation", "") == "fuzion":
                    bed = SleepIQFuzionBed(self, bed_data)
                else:
                    bed = SleepIQBed(self, bed_data)
                if await bed.valid():
                    self.beds[bed_data["bedId"]] = bed
            except SleepIQAPIException as e:
                _LOGGER.error(
                    f"Received {e.code} error setting up bed: {bed_data.get('name', 'unknown')}, skipping..."
                )

        # get sleepers and assign to beds
        data = await self.get("sleeper")
        for sleeper_data in data["sleepers"]:
            if sleeper_data["bedId"] not in self.beds:
                continue
            sleeper = self.beds[sleeper_data["bedId"]].sleepers[sleeper_data["side"]]
            sleeper.name = sleeper_data["firstName"]
            sleeper.active = sleeper_data["active"]

        # init foundations
        for bed in self.beds.values():
            await bed.foundation.fetch_features()
            await bed.foundation.init_features()

    # update statuses of sleepers/beds
    async def fetch_bed_statuses(self) -> None:
        """Update bed/sleeper statuses from API."""
        data = await self.get("bed/familyStatus")
        for bed_status in data["beds"]:
            if bed_status["bedId"] not in self.beds:
                continue
            for sleeper in self.beds[bed_status["bedId"]].sleepers:
                sleeper_data = bed_status.get(sleeper.side_full.lower() + "Side")
                if sleeper_data:
                    sleeper.in_bed = sleeper_data["isInBed"]
                    sleeper.pressure = sleeper_data["pressure"]
                    sleeper.sleep_number = sleeper_data["sleepNumber"]
                    await sleeper.update()



"""Constants for SleepIQ API Package."""
import enum

API_URL = "https://prod-api.sleepiq.sleepnumber.com/rest"
TIMEOUT = 10

LOGIN_KEY = 1
LOGIN_COOKIE = 2

RIGHT_NIGHT_STAND = 1
LEFT_NIGHT_STAND = 2
RIGHT_NIGHT_LIGHT = 3
LEFT_NIGHT_LIGHT = 4

BED_LIGHTS = [RIGHT_NIGHT_STAND, LEFT_NIGHT_STAND, RIGHT_NIGHT_LIGHT, LEFT_NIGHT_LIGHT]

class FootWarmingTemps(int, enum.Enum):
    OFF = 0
    LOW = 31
    MEDIUM = 57
    HIGH = 72

FAVORITE = 1
READ = 2
WATCH_TV = 3
FLAT = 4
ZERO_G = 5
SNORE = 6

NO_PRESET = "Not at preset"
PRESET_FAV = "Favorite"
PRESET_READ = "Read"
PRESET_TV = "Watch TV"
PRESET_FLAT = "Flat"
PRESET_ZERO_G = "Zero G"
PRESET_SNORE = "Snore"

BED_PRESETS = {
    NO_PRESET: 0,
    PRESET_FAV: FAVORITE,
    PRESET_READ: READ,
    PRESET_TV: WATCH_TV,
    PRESET_FLAT: FLAT,
    PRESET_ZERO_G: ZERO_G,
    PRESET_SNORE: SNORE,
}


class Speed(int, enum.Enum):
    OFF = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


MASSAGE_SPEEDS = [Speed.OFF, Speed.LOW, Speed.MEDIUM, Speed.HIGH]


class Mode(int, enum.Enum):
    OFF = 0
    SOOTHE = 1
    REVITILIZE = 2
    WAVE = 3


MASSAGE_MODES = [Mode.OFF, Mode.SOOTHE, Mode.REVITILIZE, Mode.WAVE]


class Side(str, enum.Enum):
    LEFT = "L"
    RIGHT = "R"
    NONE = "-"


SIDES_SHORT = {Side.LEFT: "L", Side.RIGHT: "R", Side.NONE: "R"}
SIDES_FULL = {Side.LEFT: "Left", Side.RIGHT: "Right", Side.NONE: "Right"}

FOUNDATION_TYPES = ["single", "splitHead", "splitKing", "easternKing"]


class End(str, enum.Enum):
    HEAD = "H"
    FOOT = "F"


ACTUATORS_FULL = {End.HEAD: "Head", End.FOOT: "Foot"}

BAMKEY = {
    "HaltAllActuators": "ACHA",
    "GetSystemConfiguration": "SYCG",
    "SetSleepiqPrivacyState": "SPRS",
    "GetSleepiqPrivacyState": "SPRG",
    "InterruptSleepNumberAdjustment": "PSNI",
    "StartSleepNumberAdjustment": "PSNS",
    "GetSleepNumberControls": "SNCG",
    "SetFavoriteSleepNumber": "SNFS",
    "GetFavoriteSleepNumber": "SNFG",
    "SetUnderbedLightSettings": "UBLS",
    "GetUnderbedLightSettings": "UBLG",
    "GetActuatorPosition": "ACTG",
    "SetActuatorTargetPosition": "ACTS",
    "SetTargetPresetWithoutTimer": "ASTP",
    "GetCurrentPreset": "AGCP",
}



"""Async SleepIQ API Library."""
from .asyncsleepiq import AsyncSleepIQ
from .actuator import SleepIQActuator
from .bed import SleepIQBed
from .consts import *
from .exceptions import (
    SleepIQAPIException,
    SleepIQLoginException,
    SleepIQTimeoutException,
)
from .foot_warmer import SleepIQFootWarmer
from .foundation import SleepIQFoundation
from .light import SleepIQLight
from .preset import SleepIQPreset
from .sleeper import SleepIQSleeper

__version__ = "1.4.1"



"""Actuator representation for SleepIQ API."""
from __future__ import annotations

from typing import Any

from ..actuator import SleepIQActuator


class SleepIQFuzionActuator(SleepIQActuator):
    """Actuator representation for SleepIQ API."""

    async def set_position(self, position: int, slow_speed: bool = False) -> None:
        """Set the position of an actuator through the API."""
        if position < 0 or position > 100:
            raise ValueError("Invalid position, must be between 0 and 100")
        if position == self.position:
            return
        args = [self.side_full.lower(), self.actuator_full.lower(), str(position)]
        await self._api.bamkey(self.bed_id, "SetActuatorTargetPosition", args)

    async def update(self, data: dict[str, Any]) -> None:
        """Update the position of an actuator from the API."""
        args = [self.side_full.lower(), self.actuator_full.lower()]
        result = await self._api.bamkey(self.bed_id, "GetActuatorPosition", args)
        self.position = int(result)



"""Bed object from SleepIQ API."""
from __future__ import annotations

from typing import Any

from ..api import SleepIQAPI
from ..consts import SIDES_FULL, Side
from ..foundation import SleepIQFoundation
from .foundation import SleepIQFuzionFoundation
from ..sleeper import SleepIQSleeper
from .sleeper import SleepIQFuzionSleeper
from ..bed import SleepIQBed


class SleepIQFuzionBed(SleepIQBed):
    """Fuzion Bed object from SleepIQ API."""

    def __init__(self, api: SleepIQAPI, data: dict[str, Any]) -> None:
        """Initialize bed object."""
        super().__init__(api, data)
        self.sleepers: list[SleepIQSleeper] = [
            SleepIQFuzionSleeper(api, self.id, data[f"sleeper{SIDES_FULL[side]}Id"], side)
            for side in [Side.LEFT, Side.RIGHT]
            if data.get(f"sleeper{SIDES_FULL[side]}Id")
        ]
        self.foundation: SleepIQFoundation = SleepIQFuzionFoundation(api, self.id)

    async def valid(self) -> bool:
        try:
            await self._api.bamkey(self.id, "GetSleepiqPrivacyState")
            return True
        except:
            return False

    async def stop_pump(self) -> None:
        """Stop pump."""
        await self._api.bamkey(self.id, "InterruptSleepNumberAdjustment")

    async def fetch_pause_mode(self) -> None:
        """Update paused attribute with data from API."""
        status = await self._api.bamkey(self.id, "GetSleepiqPrivacyState")
        self.paused = status == "paused"

    async def set_pause_mode(self, mode: bool) -> None:
        """Set pause mode in API and locally."""
        arg = "paused" if mode else "active"
        await self._api.bamkey(self.id, "SetSleepiqPrivacyState", args=[arg])
        self.paused = mode



from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="asyncsleepiq",
    packages=["asyncsleepiq", "asyncsleepiq.fuzion"],
    package_data={"asyncsleepiq": ["py.typed"]},
    version="1.4.1",
    description="ASync SleepIQ API",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="http://github.com/kbickar/asyncsleepiq",
    author="Keilin Bickar",
    author_email="trumpetgod@gmail.com",
    license="MIT",
    install_requires=[
        'aiohttp;python_version>="3.7"',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)



"""Light representation for SleepIQ API."""
from __future__ import annotations

from ..light import SleepIQLight


class SleepIQFuzionLight(SleepIQLight):
    """Light representation for SleepIQ API."""

    async def turn_on(self) -> None:
        """Turn on light through API."""
        await self.set_light(True)
        self.is_on = True

    async def turn_off(self) -> None:
        """Turn off light through API."""
        await self.set_light(False)
        self.is_on = False

    async def set_light(self, setting: bool) -> None:
        """Set light state through API."""
        arg = "high" if setting else "off"
        await self._api.bamkey(self.bed_id, "SetUnderbedLightSettings", args=[arg, "0"])

    async def update(self) -> None:
        """Update light state from API."""
        result = await self._api.bamkey(self.bed_id, "GetUnderbedLightSettings")
        state, num = result.split()
        self.is_on = state != "off"



"""API interface base class."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from contextlib import AbstractAsyncContextManager
import random
from typing import Any, cast

from aiohttp import ClientResponse, ClientSession, ClientTimeout

from .consts import API_URL, BAMKEY, LOGIN_KEY, TIMEOUT
from .exceptions import (
    SleepIQAPIException,
    SleepIQLoginException,
    SleepIQTimeoutException,
)


SOURCE_APP = "AsyncSleepIQ API"


def random_user_agent() -> str:
    """Create a randomly generated sorta valid User Agent string."""
    uas = {
        "Edge": ("AppleWebKit/537.36 (KHTML, like Gecko) " "Chrome/98.0.4758.80 Safari/537.36 Edg/98.0.1108.43"),
        "Chrome": ("AppleWebKit/537.36 (KHTML, like Gecko) " "Chrome/97.0.4692.99 Safari/537.36"),
        "Firefox": "Gecko/20100101 Firefox/96.0",
        "iphone": ("AppleWebKit/605.1.15 (KHTML, like Gecko) " "Version/15.2 Mobile/15E148 Safari/604.1"),
        "Safari": ("AppleWebKit/605.1.15 (KHTML, like Gecko) " "Version/11.1.2 Safari/605.1.15"),
    }
    os = {
        "windows": "Windows NT 10.0; Win64; x64",
        "iphone": "iPhone; CPU iPhone OS 15_2_1 like Mac OS X",
        "mac": "Macintosh; Intel Mac OS X 10_11_6",
    }
    template = "Mozilla/5.0 ({os}) {ua}"

    return template.format(os=random.choice(list(os.values())), ua=random.choice(list(uas.values())))


class SleepIQAPI:
    """API interface base class."""

    def __init__(
        self,
        email: str | None = None,
        password: str | None = None,
        login_method: int = LOGIN_KEY,
        client_session: ClientSession | None = None,
    ) -> None:
        """Initialize AsyncSleepIQ API Interface."""
        self.email = email
        self.password = password
        self.key = ""
        self._session = client_session or ClientSession()
        self._headers = {"User-Agent": random_user_agent()}
        self._login_method = login_method
        self._account_id = ""

    async def close_session(self) -> None:
        """Close the API session."""
        if self._session:
            await self._session.close()

    async def login(self, email: str | None = None, password: str | None = None) -> None:
        """Login using the with the email/password provided or stored."""
        if not email:
            email = self.email
        if not password:
            password = self.password
        if not email or not password:
            raise SleepIQLoginException("username/password not set")

        try:
            if self._login_method == LOGIN_KEY:
                await self.login_key(email, password)
            else:
                await self.login_cookie(email, password)

        except asyncio.TimeoutError as ex:
            # timed out
            raise SleepIQTimeoutException("API call timed out") from ex
        except SleepIQTimeoutException as ex:
            raise ex
        except Exception as ex:
            raise SleepIQLoginException(f"Connection failure: {ex}") from ex

        # store in case we need to login again
        self.email = email
        self.password = password

    async def login_key(self, email: str, password: str) -> None:
        """Login using the key authentication method with the email/password provided."""
        self.key = ""
        auth_data = {"login": email, "password": password}

        async with self._session.put(
            API_URL + "/login", headers=self._headers, timeout=TIMEOUT, json=auth_data
        ) as resp:
            if resp.status == 401:
                raise SleepIQLoginException("Incorrect username or password")
            if resp.status == 403:
                raise SleepIQLoginException("User Agent is blocked. May need to update GenUserAgent data?")
            if resp.status not in (200, 201):
                raise SleepIQLoginException(
                    "Unexpected response code: {code}\n{body}".format(
                        code=resp.status,
                        body=resp.text,
                    )
                )

            json = await resp.json()
            self.key = json["key"]

    async def login_cookie(self, email: str, password: str) -> None:
        """Login using the cookie authentication method with the email/password provided."""
        auth_data = {
            "Email": email,
            "Password": password,
            "ClientID": "2oa5825venq9kek1dnrhfp7rdh",
        }
        async with self._session.post(
            "https://l06it26kuh.execute-api.us-east-1.amazonaws.com/Prod/v1/token",
            headers=self._headers,
            timeout=TIMEOUT,
            json=auth_data,
        ) as resp:
            if resp.status == 401:
                raise SleepIQLoginException("Incorrect username or password")
            if resp.status == 403:
                raise SleepIQLoginException("User Agent is blocked. May need to update GenUserAgent data?")
            if resp.status not in (200, 201):
                raise SleepIQLoginException(
                    "Unexpected response code: {code}\n{body}".format(
                        code=resp.status,
                        body=resp.text,
                    )
                )
            json = await resp.json()
            token = json["data"]["AccessToken"]
            self._headers["Authorization"] = token

        async with self._session.get(API_URL + "/user/jwt", headers=self._headers, timeout=TIMEOUT) as resp:
            if resp.status not in (200, 201):
                raise SleepIQLoginException(
                    "Unexpected response code: {code}\n{body}".format(
                        code=resp.status,
                        body=resp.text,
                    )
                )

    async def put(self, url: str, json: dict[str, Any] = {}, params: dict[str, Any] = {}) -> dict[str, Any]:
        """Make a PUT request to the API."""
        return await self.__make_request(self._session.put, url, json, params)

    async def get(self, url: str, json: dict[str, Any] = {}, params: dict[str, Any] = {}) -> dict[str, Any] | Any:
        """Make a GET request to the API."""
        return await self.__make_request(self._session.get, url, json, params)

    async def check(self, url: str, json: dict[str, Any] = {}, params: dict[str, Any] = {}) -> bool:
        """Check if a GET request to the API would be successful."""
        return cast(
            bool,
            await self.__make_request(self._session.get, url, json, params, check=True),
        )

    async def bamkey(self, bed_id: str, key: str, args: list[str] = []) -> str:
        """Make a request to the API using the bamkey endpoint."""
        url = f"sn/v1/accounts/{self._account_id}/beds/{bed_id}/bamkey"
        json = {
            "args": " ".join(args),
            "key": BAMKEY[key],
            "sourceApplication": SOURCE_APP,
        }
        json = await self.put(url, json)
        return json.get("cdcResponse", "").replace("PASS:", "")

    async def __make_request(
        self,
        make_request: Callable[..., AbstractAsyncContextManager[ClientResponse]],
        url: str,
        json: dict[str, Any] = {},
        params: dict[str, Any] = {},
        retry: bool = True,
        check: bool = False,
    ) -> bool | dict[str, Any] | Any:
        """Make a request to the API."""
        timeout = ClientTimeout(total=TIMEOUT)
        params["_k"] = self.key
        try:
            async with make_request(
                API_URL + "/" + url,
                headers=self._headers,
                timeout=timeout,
                json=json,
                params=params,
            ) as resp:
                if check:
                    return resp.status == 200

                if resp.status != 200:
                    if retry and resp.status in (401, 404):
                        # login and try again
                        await self.login()
                        return await self.__make_request(make_request, url, json, params, False)
                    raise SleepIQAPIException(resp.status, f"API call error response {resp.status}\n{resp.text}")
                return await resp.json()
        except asyncio.TimeoutError as ex:
            # timed out
            raise SleepIQTimeoutException("API call timed out") from ex
